/******************************************************************************
*
* project name:   FAT-Engine
* file name:      fatd6.c
* initial date:   7/03/2002
* author:         thomas.nussbaumer@gmx.net
*
* description:    FAT-Engine - Demo 6 (The Legacy Intro)
*
* $Id: fatd6.c,v 1.15 2002/05/08 15:35:14 tnussb Exp $
*
******************************************************************************/

#define USE_TI89
#define USE_TI92P
#define USE_V200
#include "../includes/fat.h"

#define  MAP_WIDTH        8
#define  MAP_HEIGHT       3

char map[MAP_WIDTH*MAP_HEIGHT] = {
     1, 1, 1, 1, 1, 1, 1, 1,
     0, 0, 0, 0, 0, 0, 0, 2,
     1, 1, 1, 1, 1, 1, 1, 1};


//-----------------------------------------------------------------------------
// include the utility routines ...
//-----------------------------------------------------------------------------
#define NO_COUNTER_SUPPORT // we use no counters in this demo
#include "utils.c"

//-----------------------------------------------------------------------------
// include file containing the texture and genericdata indices into fatdtx
// (automatically generated by TexMaker)
//-----------------------------------------------------------------------------
#include "indices.h"


short display_offset = 0;    // will be set depending on calctype (TI89 or TI92p)


void WaitForMillis(unsigned short);
asm("WaitForMillis:\n"
"       movem.l %d0-%d3,-(%sp)\n"
"       move.w (20,%sp),%d2\n"
"       move.l #31,%d1\n"
"       move.l #31,%d3\n"
"_wl2_: move.w #120,%d0    /* modify this value for exact timing !!! */\n"
"_wl1_: rol.l  %d3,%d1\n"
"       dbra   %d0,_wl1_\n"
"       dbra   %d2,_wl2_\n"
"       movem.l (%sp)+,%d0-%d3\n"
"       rts");


//=============================================================================
// rotate camera and render until camera orientation final is reached
//=============================================================================
void Rotate(FATCONFIG *fc,short increment,short final) {
    do {
        FAT_Render(fc);
        fc->cam_orientation+=increment;
    }
    while(fc->cam_orientation != final);
}


//=============================================================================
// move camera and render until camera xpos final is reached
//=============================================================================
void Move(FATCONFIG *fc,short increment,short final) {
    do {
        FAT_Render(fc);
        fc->cam_xpos+=increment;
    }
    while(fc->cam_xpos != final);
    FAT_Render(fc);
}


//=============================================================================
// load "fullscreen" picture (160x100/grayscales)
//=============================================================================
void LoadPic(unsigned long* screen) {
    unsigned long* src1;
    unsigned long* src2;
    unsigned long* dest1;
    unsigned long* dest2;
    short i;

    src1 = screen;
    src2 = screen+500;
    for (i=0;i<100;i++) {
        dest1 = (unsigned long*)(GetPlane(0)+(i<<5)-(i<<1)+display_offset);
        dest2 = (unsigned long*)(GetPlane(1)+(i<<5)-(i<<1)+display_offset);
        *dest1++=*src1++,*dest1++=*src1++,*dest1++=*src1++,*dest1++=*src1++,*dest1++=*src1++;
        *dest2++=*src2++,*dest2++=*src2++,*dest2++=*src2++,*dest2++=*src2++,*dest2++=*src2++;
    }
}


//=============================================================================
// invert screen
//=============================================================================
void InvertPicture(void) {
    short i;
    unsigned long* src1 = GetPlane(0);
    unsigned long* src2 = GetPlane(1);
    unsigned long* dest1 = GetPlane(0);
    unsigned long* dest2 = GetPlane(1);

    for (i=0;i<960;i++) {
        *dest1++ = ~(*src1++),*dest2++ = ~(*src2++);
    }
}


//=============================================================================
// as usual ... our main function
//=============================================================================
void _main(void) {
    unsigned char*  bg_img = NULL;
    TEXCONFIG       textures[3];
    LCD_BUFFER      lcd;
    HANDLE          texfile  = 0;
    unsigned char*  src      = NULL;
    unsigned long*  picture1 = NULL;
    unsigned long*  picture2 = NULL;
    short           result;

    FATCONFIG fc = {
        MAP_WIDTH,  // width of map
        MAP_HEIGHT, // height of map
        map,        // pointer to map data
        textures,   // pointer to textures array
        0,          // number of sprites
        NULL,       // the sprite array
        32,         // start pos_x
        96,         // start pos_y
        96,         // orientation
        NULL,       // destination plane0
        NULL,       // destination plane1
        NULL,       // background image
        0           // frame counter
    };

    //-------------------------------------------------------------------------
    // before we try to do anything else we have to initialize the engine
    // its only valid to proceed if FAT_LOADOKAY gets returned
    //-------------------------------------------------------------------------
    if ((result = FAT_LoadEngine()) != FAT_LOADOKAY) {
        switch(result) {
            case FAT_NOTGHOSTSPACE: ST_helpMsg("ERROR: client no exepacked!");break;
            case FAT_NOTFOUND:      ST_helpMsg("ERROR: no matching engine found");break;
            case FAT_LOCKFAILED:    ST_helpMsg("ERROR: locking engine failed");break;
            case FAT_OUTOFMEM:      ST_helpMsg("ERROR: out of memory");break;
            default:                ST_helpMsg("ERROR: unknown errorcode");break;
        }
        return;
    }

    //-------------------------------------------------------------------------
    // try to load the used textures from fatdtx ...
    //-------------------------------------------------------------------------
    if ((texfile = FAT_GetHandleOfFile("fatdtx"))) {
        if ((src = FAT_LockHandleOfFile(texfile))) {
            // load textures for this demo
            if (!FAT_LoadTextures(textures,src+2,TEXIDX_BLACKWALL,3)) {
                src = 0; // just to indicate something went wrong
            }

            if (!src || !(picture1=FAT_GetGenericData(src+2,GENIDX_DEMO6SCREEN1,4000))) {
                src = 0;
            }
            if (!src || !(picture2=FAT_GetGenericData(src+2,GENIDX_DEMO6SCREEN2,4000))) {
                src = 0;
            }

        }
    }

    //-------------------------------------------------------------------------
    // check if loading failed ...
    //-------------------------------------------------------------------------
    if (!texfile || !src) {
        FAT_UnloadEngine();
        if (texfile) FAT_UnlockHandleOfFile(texfile);
        ST_helpMsg("ERROR: cannot textures from fatdtx");
        return;
    }

    LCD_save(lcd);


    //-------------------------------------------------------------------------
    // allocate space for a background image and fill the lower part of the
    // light plane with 0xff to simulate a floor ...
    //-------------------------------------------------------------------------
    if (!(bg_img = (unsigned char*)malloc(1152*2))) {
        FAT_UnloadEngine();
        FAT_UnlockHandleOfFile(texfile);
        ST_helpMsg("ERROR: out of mem");
        return;
    }
    memset(bg_img,0xff,1152*2);

    fc.background = (unsigned short*)bg_img;

    InstallDummyHandlerForINTs();

    //-------------------------------------------------------------------------
    // initialize grayscales, clear both planes and set the destination plane
    // addresses for the FAT-Engine
    //-------------------------------------------------------------------------
    if (!GrayOn()) {
        free(bg_img);
        RestoreHandlerForINTs();
        FAT_UnloadEngine();
        FAT_UnlockHandleOfFile(texfile);
        ST_helpMsg("ERROR: out of mem");
        return;
    }

    //-------------------------------------------------------------------------
    // display the FAT-Engine's splashscreen
    //-------------------------------------------------------------------------
    FAT_DisplayLogo(100);

    memset(GetPlane(0),0xff,LCD_SIZE);
    memset(GetPlane(1),0xff,LCD_SIZE);

    fc.dest_plane0 = GetPlane(0)+4;
    fc.dest_plane1 = GetPlane(1)+4;

    //-------------------------------------------------------------------------
    // setup calculator type specials
    //-------------------------------------------------------------------------
    if (TI89)         display_offset = 0;
    else              display_offset = (14*30+4);

    fc.dest_plane0  += display_offset>>1;
    fc.dest_plane1  += display_offset>>1;


    //-------------------------------------------------------------------------
    // start the intro ...
    //-------------------------------------------------------------------------
    Rotate(&fc,-16,0);
    Move(&fc,16,6*64);
    WaitForMillis(3000);
    Move(&fc,-16,96);
    Rotate(&fc,-16,-ANGLE_090);

    fc.cam_xpos = 32;
    fc.cam_ypos = 96;
    map[15]     = 3;

    Rotate(&fc,16,0);
    Move(&fc,16,6*64);
    WaitForMillis(4000);
    Move(&fc,-16,96);
    Rotate(&fc,16,ANGLE_090);

    LoadPic(picture1); //InvertPicture();
    WaitForMillis(3000);

    while (!FAT_GetPadState()) {
        if (random(100)<80) LoadPic(picture1); //InvertPicture();
        else                LoadPic(picture2);
        WaitForMillis(random(2000));
        InvertPicture();
        WaitForMillis(random(300));
        InvertPicture();
    }


    //-------------------------------------------------------------------------
    // some clean stuff ...
    //-------------------------------------------------------------------------
    FAT_AllKeysReleased();
    GrayOff();
    free(bg_img);
    LCD_restore(lcd);
    ST_helpMsg(FAT_GetPoweredByString());
    FAT_UnlockHandleOfFile(texfile);  // unlock texture file
    FAT_UnloadEngine();
    RestoreHandlerForINTs();
    OSqclear(kbd_queue());
}

//#############################################################################
// Revision History
//#############################################################################
//
// $Log: fatd6.c,v $
// Revision 1.15  2002/05/08 15:35:14  tnussb
// (1) display of FAT splashscreen (FAT_DisplayLogo) added
// (2) GUI of demos completely revised
//
// Revision 1.14  2002/05/07 15:22:13  tnussb
// stupid double call of FAT_LoadEngine() fixed
//
// Revision 1.13  2002/04/30 17:41:44  tnussb
// using now defines from automatically generated headerfile indices.h for indices into
// our texture file
//
// Revision 1.12  2002/04/26 13:10:49  tnussb
// if FAT_LoadEngine() fails different strings will be displayed yet
//
// Revision 1.11  2002/04/25 19:33:15  tnussb
// generic commit for engine version v1.05
//
// Revision 1.10  2002/04/24 17:42:51  tnussb
// (1) using now FAT_GetPadState() to query key presses
//
// Revision 1.9  2002/04/18 20:27:37  tnussb
// loading now all graphical data from file fatdtx
//
// Revision 1.8  2002/04/18 12:14:47  tnussb
// using now fatdtx texture file instead of embedded textures
//
// Revision 1.7  2002/04/16 11:36:37  tnussb
// All calls to GrayMode() replaced by GrayOn/GrayOff calls
//
// Revision 1.6  2002/04/15 10:00:12  tnussb
// using now STRIP_NORMAL instead of 0 for mirrored field structure TEXSTRIP
//
// Revision 1.5  2002/04/08 20:18:16  tnussb
// using now signed char (instead of unsigned char) for map
//
// Revision 1.4  2002/03/26 11:56:24  tnussb
// all warnings reported by -Wall fixed
//
// Revision 1.3  2002/03/11 09:17:43  tnussb
// USE_V200 added
//
// Revision 1.2  2002/03/08 13:37:27  tnussb
// missing LCD_restore() added
//
// Revision 1.1  2002/03/07 11:12:41  tnussb
// initial version
//
//
